"""
AA-Forum url config
"""

from django.conf.urls import url

from aa_forum.views import admin, forum

app_name: str = "aa_forum"

urlpatterns = [
    url(r"^$", forum.index, name="forum_index"),
    # Admin URLs (These need to be first!)
    url(r"^admin/$", admin.index, name="admin_index"),
    url(
        r"^admin/category/create/$",
        admin.category_create,
        name="admin_category_create",
    ),
    url(
        r"^admin/category/(?P<category_id>[0-9]+)/edit/$",
        admin.category_edit,
        name="admin_category_edit",
    ),
    url(
        r"^admin/category/(?P<category_id>[0-9]+)/delete/$",
        admin.category_delete,
        name="admin_category_delete",
    ),
    url(
        r"^admin/category/(?P<category_id>[0-9]+)/board/create/$",
        admin.board_create,
        name="admin_board_create",
    ),
    url(
        r"^admin/category/(?P<category_id>[0-9]+)/board/(?P<board_id>[0-9]+)/edit/$",
        admin.board_edit,
        name="admin_board_edit",
    ),
    url(
        r"^admin/category/(?P<category_id>[0-9]+)/board/(?P<board_id>[0-9]+)/delete/$",
        admin.board_delete,
        name="admin_board_delete",
    ),
    url(
        r"^ajax/admin/category-order/$",
        admin.ajax_category_order,
        name="admin_ajax_category_order",
    ),
    url(
        r"^ajax/admin/board-order/$",
        admin.ajax_board_order,
        name="admin_ajax_board_order",
    ),
    # Service URLs (These have to be before the forum URLs!)
    url(
        r"^message/(?P<message_id>[0-9]+)/$",
        forum.message_entry_point_in_topic,
        name="forum_message_entry_point_in_topic",
    ),
    url(
        r"^message/(?P<message_id>[0-9]+)/delete/$",
        forum.message_delete,
        name="forum_message_delete",
    ),
    url(
        r"^topic/(?P<topic_id>[0-9]+)/change-lock-state/$",
        forum.topic_change_lock_state,
        name="forum_topic_change_lock_state",
    ),
    url(
        r"^topic/(?P<topic_id>[0-9]+)/change-sticky-state/$",
        forum.topic_change_sticky_state,
        name="forum_topic_change_sticky_state",
    ),
    url(
        r"^topic/(?P<topic_id>[0-9]+)/delete/$",
        forum.topic_delete,
        name="forum_topic_delete",
    ),
    # Forum URLs (This needs to be the last block!)
    url(
        r"^(?P<category_slug>[\w-]+)/(?P<board_slug>[\w-]+)/$",
        forum.board,
        name="forum_board",
    ),
    url(
        r"^(?P<category_slug>[\w-]+)/(?P<board_slug>[\w-]+)/(?P<page_number>[0-9]+)/$",
        forum.board,
        name="forum_board",
    ),
    url(
        r"^(?P<category_slug>[\w-]+)/(?P<board_slug>[\w-]+)/new-topic/$",
        forum.board_new_topic,
        name="forum_board_new_topic",
    ),
    url(
        r"^(?P<category_slug>[\w-]+)/(?P<board_slug>[\w-]+)/(?P<topic_slug>[\w-]+)/$",
        forum.topic,
        name="forum_topic",
    ),
    url(
        r"^(?P<category_slug>[\w-]+)/(?P<board_slug>[\w-]+)/"
        r"(?P<topic_slug>[\w-]+)/(?P<page_number>[0-9]+)/$",
        forum.topic,
        name="forum_topic",
    ),
    url(
        r"^(?P<category_slug>[\w-]+)/(?P<board_slug>[\w-]+)/"
        r"(?P<topic_slug>[\w-]+)/reply/$",
        forum.topic_reply,
        name="forum_topic_reply",
    ),
    url(
        r"^(?P<category_slug>[\w-]+)/(?P<board_slug>[\w-]+)/"
        r"(?P<topic_slug>[\w-]+)/modify/(?P<message_id>[\w-]+)/$",
        forum.message_modify,
        name="forum_message_modify",
    ),
]



# Generated by Django 3.2.4 on 2021-06-10 13:23

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

import aa_forum.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "verbose_name": "AA-Forum",
                "permissions": (
                    ("basic_access", "Can access the AA-Forum module"),
                    (
                        "manage_forum",
                        "Can manage the AA-Forum module (Categories, topics and "
                        "messages)",
                    ),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Boards",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(default="", max_length=254)),
                ("description", models.TextField(blank=True, null=True)),
                ("order", models.IntegerField(default=0)),
            ],
            options={
                "verbose_name": "Board",
                "verbose_name_plural": "Boards",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Settings",
            fields=[
                (
                    "variable",
                    models.CharField(
                        max_length=254, primary_key=True, serialize=False, unique=True
                    ),
                ),
                ("value", models.TextField()),
            ],
            options={
                "verbose_name": "Setting",
                "verbose_name_plural": "Settings",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Slugs",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("slug", models.SlugField(allow_unicode=True, max_length=255)),
            ],
            options={
                "verbose_name": "Slug",
                "verbose_name_plural": "Slugs",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Topics",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("subject", models.CharField(default="", max_length=254)),
                ("is_sticky", models.BooleanField(db_index=True, default=False)),
                ("is_locked", models.BooleanField(db_index=True, default=False)),
                ("num_views", models.IntegerField(default=0)),
                ("num_replies", models.IntegerField(default=0)),
                (
                    "time_modified",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                (
                    "board",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="topics",
                        to="aa_forum.boards",
                    ),
                ),
                (
                    "read_by",
                    models.ManyToManyField(
                        blank=True,
                        related_name="aa_forum_read_topics",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "slug",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="aa_forum.slugs",
                    ),
                ),
                (
                    "user_started",
                    models.ForeignKey(
                        on_delete=models.SET(aa_forum.models.get_sentinel_user),
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "user_updated",
                    models.ForeignKey(
                        on_delete=models.SET(aa_forum.models.get_sentinel_user),
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "Topic",
                "verbose_name_plural": "Topics",
                "ordering": ["-time_modified"],
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="PersonalMessages",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("time_sent", models.DateTimeField(default=django.utils.timezone.now)),
                ("subject", models.CharField(default="", max_length=254)),
                ("message", models.TextField(blank=True, null=True)),
                ("is_read", models.BooleanField(default=False)),
                (
                    "recipient",
                    models.ForeignKey(
                        on_delete=models.SET(aa_forum.models.get_sentinel_user),
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "sender",
                    models.ForeignKey(
                        on_delete=models.SET(aa_forum.models.get_sentinel_user),
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "slug",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="aa_forum.slugs",
                    ),
                ),
            ],
            options={
                "verbose_name": "Personal Message",
                "verbose_name_plural": "Personal Messages",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Messages",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "time_posted",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                (
                    "time_modified",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("message", models.TextField(blank=True, null=True)),
                (
                    "board",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="messages",
                        to="aa_forum.boards",
                    ),
                ),
                (
                    "read_by",
                    models.ManyToManyField(
                        blank=True,
                        related_name="aa_forum_read_messages",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "topic",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="messages",
                        to="aa_forum.topics",
                    ),
                ),
                (
                    "user_created",
                    models.ForeignKey(
                        on_delete=models.SET(aa_forum.models.get_sentinel_user),
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "user_updated",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=models.SET(aa_forum.models.get_sentinel_user),
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "Message",
                "verbose_name_plural": "Messages",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Categories",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(default="", max_length=254)),
                ("is_collapsible", models.BooleanField(default=False)),
                ("order", models.IntegerField(default=0)),
                (
                    "slug",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="aa_forum.slugs",
                    ),
                ),
            ],
            options={
                "verbose_name": "Category",
                "verbose_name_plural": "Categories",
                "default_permissions": (),
            },
        ),
        migrations.AddField(
            model_name="boards",
            name="category",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="boards",
                to="aa_forum.categories",
            ),
        ),
        migrations.AddField(
            model_name="boards",
            name="groups",
            field=models.ManyToManyField(
                blank=True,
                related_name="aa_forum_boards_group_restriction",
                to="auth.Group",
            ),
        ),
        migrations.AddField(
            model_name="boards",
            name="parent_board",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="child_boards",
                to="aa_forum.boards",
            ),
        ),
        migrations.AddField(
            model_name="boards",
            name="slug",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="aa_forum.slugs",
            ),
        ),
    ]



"""
Forum related views
"""

import math

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required

# from django.core import serializers
from django.core.handlers.wsgi import WSGIRequest
from django.core.paginator import Paginator
from django.db.models import Count, Prefetch, Q
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

from aa_forum.constants import SETTING_MESSAGESPERPAGE, SETTING_TOPICSPERPAGE
from aa_forum.forms import EditMessageForm, NewTopicForm
from aa_forum.models import Boards, Categories, Messages, Settings, Topics

# from aa_forum.tasks import set_messages_read_by_user_in_pagination


@login_required
@permission_required("aa_forum.basic_access")
def index(request: WSGIRequest) -> HttpResponse:
    """
    Forum index view
    :param request:
    :type request:
    :return:
    :rtype:
    """

    categories = (
        Categories.objects.prefetch_related(
            Prefetch(
                "boards",
                queryset=Boards.objects.prefetch_related("messages")
                .filter(
                    Q(groups__in=request.user.groups.all()) | Q(groups__isnull=True),
                    parent_board__isnull=True,
                )
                .distinct()
                .annotate(
                    num_posts=Count("messages", distinct=True),
                    num_topics=Count("topics", distinct=True),
                )
                .order_by("order"),
            )
        )
        .all()
        .order_by("order")
    )

    context = {"categories": categories}

    return render(request, "aa_forum/view/forum/index.html", context)


@login_required
@permission_required("aa_forum.basic_access")
def board(
    request: WSGIRequest, category_slug: str, board_slug: str, page_number: int = None
) -> HttpResponse:
    """
    Forum board view
    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :return:
    :rtype:
    """

    try:
        board = (
            Boards.objects.prefetch_related("messages")
            # .prefetch_related("child_boards")
            .prefetch_related(
                Prefetch(
                    "topics",
                    queryset=Topics.objects.prefetch_related("messages")
                    .distinct()
                    .annotate(
                        num_posts=Count("messages", distinct=True),
                    )
                    .order_by("-is_sticky", "-time_modified"),
                )
            )
            .filter(
                Q(groups__in=request.user.groups.all()) | Q(groups__isnull=True),
                category__slug__slug__exact=category_slug,
                slug__slug__exact=board_slug,
            )
            .distinct()
            .get()
        )
    except Boards.DoesNotExist:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>The board you were trying to visit does "
                    "either not exist, or you don't have access to it ...</p>"
                )
            ),
        )

        return redirect("aa_forum:forum_index")

    paginator = Paginator(
        board.topics.all(),
        int(Settings.objects.get_setting(setting_key=SETTING_TOPICSPERPAGE)),
    )
    page_obj = paginator.get_page(page_number)

    context = {"board": board, "page_obj": page_obj}

    return render(request, "aa_forum/view/forum/board.html", context)


@login_required
@permission_required("aa_forum.basic_access")
def board_new_topic(
    request: WSGIRequest, category_slug: str, board_slug: str
) -> HttpResponse:
    """
    Beginn a new topic
    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :return:
    :rtype:
    """

    try:
        Categories.objects.get(slug__slug__exact=category_slug)
    except Categories.DoesNotExist:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>The category you were trying to post in does "
                    "not exist ...</p>"
                )
            ),
        )

        return redirect("aa_forum:forum_index")

    try:
        board = (
            Boards.objects.filter(
                Q(groups__in=request.user.groups.all()) | Q(groups__isnull=True),
                category__slug__slug__exact=category_slug,
                slug__slug__exact=board_slug,
            )
            .distinct()
            .get()
        )
    except Boards.DoesNotExist:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>The board you were trying to post in does "
                    "either not exist, or you don't have access to it ...</p>"
                )
            ),
        )

        return redirect("aa_forum:forum_index")

    # If this is a POST request we need to process the form data
    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = NewTopicForm(request.POST)

        # Check whether it's valid:
        if form.is_valid():
            user_started = request.user
            user_updated = request.user
            post_time = timezone.now()

            topic = Topics()
            topic.board = board
            topic.user_started = user_started
            topic.user_updated = user_updated
            topic.time_modified = post_time
            topic.subject = form.cleaned_data["subject"]
            topic.save()

            message = Messages()
            message.topic = topic
            message.board = board
            message.time_posted = post_time
            message.time_modified = post_time
            message.user_created = user_started
            message.message = form.cleaned_data["message"]
            message.save()

            # Set topic and message as "read by" by the author
            topic.read_by.add(user_updated)
            message.read_by.add(request.user)

            return redirect(
                "aa_forum:forum_board",
                category_slug=board.category.slug.slug,
                board_slug=board.slug.slug,
            )

    # If not, we'll create a blank form
    else:
        form = NewTopicForm()

    context = {"board": board, "form": form}

    return render(request, "aa_forum/view/forum/new-topic.html", context)


@login_required
@permission_required("aa_forum.basic_access")
def topic(
    request: WSGIRequest,
    category_slug: str,
    board_slug: str,
    topic_slug: str,
    page_number: int = None,
) -> HttpResponse:
    """
    View a topic
    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :param page_number:
    :type page_number:
    :return:
    :rtype:
    """

    try:
        Boards.objects.prefetch_related("messages").filter(
            Q(groups__in=request.user.groups.all()) | Q(groups__isnull=True),
            category__slug__slug__exact=category_slug,
            slug__slug__exact=board_slug,
        ).distinct().get()
    except Boards.DoesNotExist:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>The topic you were trying to view does "
                    "either not exist, or you don't have access to it ...</p>"
                )
            ),
        )

        return redirect("aa_forum:forum_index")

    try:
        topic = Topics.objects.get(slug__slug__exact=topic_slug)
    except Topics.DoesNotExist:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>The topic you were trying to view does not "
                    "exist ...</p>"
                )
            ),
        )

        return redirect("aa_forum:forum_index")

    topic_messages = Messages.objects.filter(topic=topic)

    # Set this topic as "read by" by the current user
    topic.read_by.add(request.user)

    paginator = Paginator(
        topic_messages,
        int(Settings.objects.get_setting(setting_key=SETTING_MESSAGESPERPAGE)),
    )
    page_obj = paginator.get_page(page_number)

    # Set the messages as "read by" the current user
    # MessagesReadByUsers = Messages.read_by.through
    # MessagesReadByUsers.objects.bulk_create(
    #     [
    #         MessagesReadByUsers(messages_id=pk, user=request.user)
    #         for pk in page_obj.object_list.values_list("pk", flat=True)
    #     ]
    # )

    # messages_this_page = page_obj.object_list
    # last_message = messages_this_page[
    #     int(Settings.objects.get_setting(setting_key=SETTING_MESSAGESPERPAGE)) - 1
    # ]

    # request.user.aa_forum_read_messages.add(*page_obj.object_list)

    # set_messages_read_by_user_in_pagination.delay(
    #     object_list=serializers.serialize("json", page_obj.object_list),
    #     user_id=request.user.id,
    # )

    reply_form = EditMessageForm()

    context = {
        "topic": topic,
        "page_obj": page_obj,
        "reply_form": reply_form,
        # "messages_this_page": messages_this_page,
        # "last_message": last_message.id,
    }

    return render(request, "aa_forum/view/forum/topic.html", context)


@login_required
@permission_required("aa_forum.basic_access")
def topic_reply(
    request: WSGIRequest, category_slug: str, board_slug: str, topic_slug: str
) -> HttpResponse:
    """
    Reply to posts in a topic
    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :return:
    :rtype:
    """

    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditMessageForm(request.POST)

        # Check whether it's valid:
        if form.is_valid():
            board = Boards.objects.get(slug__slug__exact=board_slug)
            topic = Topics.objects.get(slug__slug__exact=topic_slug)
            time_posted = timezone.now()

            new_message = Messages()
            new_message.topic = topic
            new_message.board = board
            new_message.user_created = request.user
            new_message.time_posted = time_posted
            new_message.time_modified = time_posted
            new_message.message = form.cleaned_data["message"]
            new_message.save()

            # Update the modified timestamp on the topic
            topic.time_modified = time_posted
            topic.save()

            # Remove all users from "read by" list and set the current user again.
            # This way we mark this topic as unread for all but the current user.
            topic.read_by.clear()
            topic.read_by.add(request.user)

            # Set the message as "read by" the author
            new_message.read_by.add(request.user)

            return redirect(
                "aa_forum:forum_message_entry_point_in_topic", new_message.id
            )

    messages.warning(
        request,
        mark_safe(_("<h4>Warning!</h4><p>Something went wrong, please try again</p>.")),
    )

    return redirect("aa_forum:forum_topic", category_slug, board_slug, topic_slug)


@login_required
@permission_required("aa_forum.manage_forum")
def topic_change_lock_state(
    request: WSGIRequest, topic_id: int
) -> HttpResponseRedirect:
    """
    Change the lock state of the given topic
    :param request:
    :type request:
    :param topic_id:
    :type topic_id:
    :return:
    :rtype:
    """

    topic = Topics.objects.get(pk=topic_id)

    if topic.is_locked:
        topic.is_locked = False

        messages.success(
            request,
            mark_safe(_("<h4>Success!</h4><p>Topic has been unlocked/re-opened.</p>")),
        )
    else:
        topic.is_locked = True

        messages.success(
            request,
            mark_safe(_("<h4>Success!</h4><p>Topic has been locked/closed.</p>")),
        )

    topic.save()

    return redirect("aa_forum:forum_board", topic.board.category.slug, topic.board.slug)


@login_required
@permission_required("aa_forum.manage_forum")
def topic_change_sticky_state(
    request: WSGIRequest, topic_id: int
) -> HttpResponseRedirect:
    """
    Change the sticky state of the given topic
    :param request:
    :type request:
    :param topic_id:
    :type topic_id:
    :return:
    :rtype:
    """

    topic = Topics.objects.get(pk=topic_id)

    if topic.is_sticky:
        topic.is_sticky = False

        messages.success(
            request,
            mark_safe(_('<h4>Success!</h4><p>Topic is no longer "Sticky".</p>')),
        )
    else:
        topic.is_sticky = True

        messages.success(
            request,
            mark_safe(_('<h4>Success!</h4><p>Topic is now "Sticky".</p>')),
        )

    topic.save()

    return redirect("aa_forum:forum_board", topic.board.category.slug, topic.board.slug)


@login_required
@permission_required("aa_forum.manage_forum")
def topic_delete(request: WSGIRequest, topic_id: int) -> HttpResponseRedirect:
    """
    Delete a given topic
    :param request:
    :type request:
    :param topic_id:
    :type topic_id:
    """

    topic = Topics.objects.get(pk=topic_id)
    board = topic.board

    topic.delete()

    messages.success(
        request,
        mark_safe(_("<h4>Success!</h4><p>Topic removed.</p>")),
    )

    return redirect("aa_forum:forum_board", board.category.slug, board.slug)


@login_required
@permission_required("aa_forum.basic_access")
def message_entry_point_in_topic(
    request: WSGIRequest, message_id: int
) -> HttpResponseRedirect:
    """
    Get a messages antry point in a topic, so we end up on the right page with it
    :param request:
    :type request:
    :param message_id:
    :type message_id:
    """

    try:
        message = Messages.objects.get(pk=message_id)
    except Messages.DoesNotExist:
        messages.error(
            request,
            mark_safe(_("<h4>Error!</h4><p>The message doesn't exist ...</p>")),
        )

        return redirect("aa_forum:forum_index")

    try:
        board = (
            Boards.objects.filter(
                Q(groups__in=request.user.groups.all()) | Q(groups__isnull=True),
                pk=message.board.pk,
            )
            .distinct()
            .get()
        )
    except Boards.DoesNotExist:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>The topic you were trying to view does "
                    "either not exist, or you don't have access to it ...</p>"
                )
            ),
        )

        return redirect("aa_forum:forum_index")

    messages_in_topic = Messages.objects.filter(pk__lte=message.pk, topic=message.topic)
    number_of_messages_in_topic = messages_in_topic.count()
    settings = Settings.objects.all()
    messages_per_topic = settings.values_list("value", flat=True).get(
        variable__exact="defaultMaxMessages"
    )

    page = int(math.ceil(int(number_of_messages_in_topic) / int(messages_per_topic)))

    if page > 1:
        redirect_path = reverse(
            "aa_forum:forum_topic",
            args=(
                board.category.slug,
                board.slug,
                message.topic.slug,
                page,
            ),
        )
        redirect_url = f"{redirect_path}#message-{message.pk}"
    else:
        redirect_path = reverse(
            "aa_forum:forum_topic",
            args=(board.category.slug, board.slug, message.topic.slug),
        )
        redirect_url = f"{redirect_path}#message-{message.pk}"

    return HttpResponseRedirect(redirect_url)


@login_required
@permission_required("aa_forum.basic_access")
def message_modify(
    request: WSGIRequest,
    category_slug: str,
    board_slug: str,
    topic_slug: str,
    message_id: int,
) -> HttpResponse:
    """
    Modify a given message
    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :param message_id:
    :type message_id:
    """

    # Check if the user has access to this board in the first place
    try:
        board = (
            Boards.objects.filter(
                Q(groups__in=request.user.groups.all()) | Q(groups__isnull=True),
                slug__slug__exact=board_slug,
            )
            .distinct()
            .get()
        )
    except Boards.DoesNotExist:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>The topic you were trying to view does "
                    "either not exist, or you don't have access to it ...</p>"
                )
            ),
        )

        return redirect("aa_forum:forum_index")

    # If the user has access, check if the message exists
    try:
        message = Messages.objects.get(pk=message_id)
    except Messages.DoesNotExist:
        messages.error(
            request,
            mark_safe(_("<h4>Error!</h4><p>The message doesn't exist ...</p>")),
        )

        return redirect("aa_forum:forum_index")

    # Check if the user actually has the right to edit this message
    if message.user_created_id is not request.user.id and not request.user.has_perm(
        "aa_forum.manage_forum"
    ):
        messages.error(
            request,
            mark_safe(
                _("<h4>Error!</h4><p>You are not allowed to modify this message!</p>")
            ),
        )

        return redirect("aa_forum:forum_index")

    # We are in the clear, let's see what we've got
    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditMessageForm(request.POST)

        # Check whether it's valid:
        if form.is_valid():
            user_updated = request.user
            updated_time = timezone.now()

            # topic = Topics.objects.get(slug__slug__exact=topic_slug)
            # topic.time_modified = updated_time
            # topic.save()

            message.user_updated = user_updated
            message.time_modified = updated_time
            message.message = form.cleaned_data["message"]
            message.save()

            messages.success(
                request,
                mark_safe(_("<h4>Success!</h4><p>The message has been updated.</p>")),
            )

            return redirect(
                "aa_forum:forum_message_entry_point_in_topic", message_id=message_id
            )

    # If not, we'll fill the form with the information from the message object
    else:
        form = EditMessageForm(instance=message)

    context = {"form": form, "board": board, "message": message}

    return render(request, "aa_forum/view/forum/modify-message.html", context)


@login_required
@permission_required("aa_forum.manage_forum")
def message_delete(request: WSGIRequest, message_id: int) -> HttpResponseRedirect:
    """
    Delete a message from a topic
    If it is the last message in this topic, the topic will be removed as well
    :param request:
    :type request:
    :param message_id:
    :type message_id:
    """

    message = Messages.objects.get(pk=message_id)
    topic = message.topic

    # Let's check if we have more than one message in this topic
    # If so, remove just that message and return to the topic
    if topic.messages.all().count() > 1:
        message.delete()

        messages.success(
            request,
            mark_safe(_("<h4>Success!</h4><p>The message has been deleted.</p>")),
        )

        return redirect(
            "aa_forum:forum_topic",
            category_slug=topic.board.category.slug,
            board_slug=topic.board.slug,
            topic_slug=topic.slug,
        )

    # If it is the only/last message in the topic, remove the topic
    topic.delete()

    messages.success(
        request,
        mark_safe(
            _(
                "<h4>Success!</h4><p>The message has been deleted.</p><p>This was "
                "the only/last message in this topic, so the topic has been "
                "removed as well.</p>"
            )
        ),
    )

    return redirect(
        "aa_forum:forum_board",
        category_slug=topic.board.category.slug,
        board_slug=topic.board.slug,
    )



"""
Fake some messages
"""

from faker import Faker

from django.utils import timezone

from aa_forum.models import Messages, Topics


def run():
    """
    Create some random messages for topic id 1
    """

    fake = Faker()

    topics = Topics.objects.all()
    time_posted = timezone.now()

    if topics.count() > 0:
        for topic in topics:
            Messages.objects.bulk_create(
                [
                    Messages(
                        message=fake.sentence(),
                        time_posted=time_posted,
                        time_modified=time_posted,
                        board_id=topic.board_id,
                        topic_id=topic.id,
                        user_created_id=2,
                    )
                    for _ in range(25)
                ]
            )



"""
We need some tasks, so here we are ...
"""

from celery import shared_task

from django.contrib.auth.models import User

# from django.core import serializers
# from django.core.paginator import Page

# from aa_forum.models import Messages


@shared_task
def set_messages_read_by_user_in_pagination(object_list, user_id: int):
    """
    Set messages read by user in a pagination context
    :param page_object:
    :type page_object:
    :param user:
    :type user:
    """

    # MessagesReadByUsers = Messages.read_by.through
    # MessagesReadByUsers.objects.bulk_create(
    #     [
    #         MessagesReadByUsers(messages_id=pk, user=user)
    #         for pk in page_object.object_list.values_list("pk", flat=True)
    #     ]
    # )

    user = User.objects.get(pk=user_id)

    # user.aa_forum_read_messages.add(*serializers.deserialize("json", object_list))
    user.aa_forum_read_messages.add(*object_list)



"""
Forms
"""

from ckeditor_uploader.widgets import CKEditorUploadingWidget

from django import forms
from django.contrib.auth.models import Group
from django.forms import ModelForm
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from aa_forum.models import Boards, Categories, Messages


def get_mandatory_form_label_text(text):
    """
    Label text for mandatory form fields
    :param text:
    :type text:
    :return:
    :rtype:
    """

    required_text = _("This field is mandatory")
    required_marker = (
        f'<span aria-label="{required_text}" class="form-required-marker">*</span>'
    )

    return mark_safe(
        f'<span class="form-field-required">{text} {required_marker}</span>'
    )


class NewTopicForm(forms.Form):
    """
    New topic form
    """

    subject = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(_("Subject")),
        max_length=255,
        widget=forms.TextInput(attrs={"placeholder": _("Subject")}),
    )

    message = forms.CharField(
        widget=CKEditorUploadingWidget(
            attrs={"rows": 10, "cols": 20, "style": "width: 100%;"}
        ),
        required=True,
        label=get_mandatory_form_label_text(_("Message")),
    )


class EditCategoryForm(ModelForm):
    """
    New category form
    """

    name = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(_("Name")),
        max_length=255,
        widget=forms.TextInput(attrs={"placeholder": _("Category Name")}),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Categories
        fields = ["name"]


class EditBoardForm(ModelForm):
    """
    New board form
    """

    name = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(_("Name")),
        max_length=255,
        widget=forms.TextInput(attrs={"placeholder": _("Board Name")}),
    )
    description = forms.CharField(
        required=False,
        label=_("Description"),
        widget=forms.Textarea(
            attrs={
                "rows": 10,
                "cols": 20,
                "input_type": "textarea",
                "placeholder": _("Board Description (Optional)"),
            }
        ),
    )
    groups = forms.ModelMultipleChoiceField(
        required=False, queryset=Group.objects.all()
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Boards
        fields = ["name", "description", "groups"]


class EditMessageForm(ModelForm):
    """
    Reply form
    """

    message = forms.CharField(
        widget=CKEditorUploadingWidget(
            attrs={"rows": 10, "cols": 20, "style": "width: 100%;"}
        ),
        required=True,
        label=get_mandatory_form_label_text(_("Message")),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Messages
        fields = ["message"]



"""
Constants
"""

from django.utils.text import slugify

from aa_forum import __version__

github_url = "https://github.com/ppfeufer/aa-forum"
verbose_name = "AA-Forum - A simple forum for Alliance Auth"
verbose_name_slug = slugify(verbose_name, allow_unicode=True)
user_agent = f"{verbose_name_slug} v{__version__} {github_url}"

# Settings keys
SETTING_DATEFORMAT = "defaultDateFormat"
SETTING_TIMEFORMAT = "defaultTimeFormat"
SETTING_MAXMESSAGELENGTH = "maxMessageLength"
SETTING_MESSAGESPERPAGE = "defaultMaxMessages"
SETTING_TOPICSPERPAGE = "defaultMaxTopics"
SETTING_OLDTOPICDAYS = "oldTopicDays"



# Generated by Django 3.2.4 on 2021-06-09 13:10

from django.db import migrations

reserved_slugs_to_migrate = [
    {"slug": "admin"},
    {"slug": "administrator"},
    {"slug": "administration"},
    {"slug": "category-order"},
    {"slug": "board-order"},
]


def add_reserved_slugs(apps, schema_editor):
    """
    Add reserved slugs on migration
    :param apps:
    :type apps:
    :param schema_editor:
    :type schema_editor:
    """

    Slugs = apps.get_model("aa_forum", "Slugs")
    db_alias = schema_editor.connection.alias

    reserved_slugs = []
    for reserved_slug in reserved_slugs_to_migrate:
        reserved_slugs.append(Slugs(slug=reserved_slug["slug"]))

    Slugs.objects.using(db_alias).bulk_create(reserved_slugs)


def remove_reserved_slugs(apps, schema_editor):
    """
    Remove reserved slugs on migration to zero
    :param apps:
    :type apps:
    :param schema_editor:
    :type schema_editor:
    """

    Slugs = apps.get_model("aa_forum", "Slugs")
    db_alias = schema_editor.connection.alias

    for reserved_slug in reserved_slugs_to_migrate:
        Slugs.objects.using(db_alias).filter(slug=reserved_slug["slug"]).delete()


class Migration(migrations.Migration):

    dependencies = [
        ("aa_forum", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(add_reserved_slugs, remove_reserved_slugs),
    ]



"""
Our app setting
"""

from aasrp.utils import clean_setting

from django.apps import apps

# AA-GDPR
AVOID_CDN = clean_setting("AVOID_CDN", False)


def avoid_cdn() -> bool:
    """
    Check if we should avoid CDN usage
    :return: bool
    """

    return AVOID_CDN


def allianceauth_discordbot_active():
    """
    Check if allianceauth-discordbot is installed and active
    :return:
    """

    return apps.is_installed("aadiscordbot")


def aa_discordnotify_active():
    """
    Check if aa-discordnotfify is installed and active
    :return:
    """

    return apps.is_installed("discordnotify")
